# Copyright (c) 2025 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

from __future__ import annotations

import collections.abc
import math
import typing

from ....compat import import_optional_dependency
from ....time import TimeUnit
from ..record import MessagePathRecord
from .timestamp import Timestamp

if typing.TYPE_CHECKING:
    import pyarrow  # pants: no-infer-dep
    import pyarrow.parquet  # pants: no-infer-dep


def compute_time_filter_mask(
    timestamps: "pyarrow.Array",
    start_time: typing.Optional[int] = None,
    end_time: typing.Optional[int] = None,
) -> typing.Optional["pyarrow.BooleanArray"]:
    """
    Compute a boolean mask indicating which rows fall within the specified time range.
    Returns None if no time filtering is needed (both start_time and end_time are None).
    """
    pc = import_optional_dependency("pyarrow.compute", "analytics")

    if start_time is None and end_time is None:
        return None

    masks: list["pyarrow.BooleanArray"] = []
    if start_time is not None:
        # Temporal filtering is inclusive of start time
        masks.append(pc.greater_equal(timestamps, start_time))

    if end_time is not None:
        # Temporal filtering is exclusive of end time
        masks.append(pc.less(timestamps, end_time))

    if len(masks) == 1:
        return masks[0]

    # Combine masks using Kleene logic (a null value means “unknown”, and an unknown value ‘and’ false is always false)
    return pc.and_kleene(masks[0], masks[1])


def extract_timestamps(
    table: "pyarrow.Table",
    timestamp: Timestamp,
) -> "pyarrow.Int64Array":
    """
    Extract timestamps in nanoseconds since Unix epoch from the table's timestamp column.
    """
    pa = import_optional_dependency("pyarrow", "analytics")
    pc = import_optional_dependency("pyarrow.compute", "analytics")

    timestamp_values: "pyarrow.Array" = table.column(timestamp.field.name).combine_chunks()
    data_type = timestamp.field.type

    # Short-circuit if timestamp message path column is already in target format/unit
    if pa.types.is_int64(data_type) and timestamp.unit() == TimeUnit.Nanoseconds:
        return typing.cast("pyarrow.Int64Array", timestamp_values)

    # Derive nanoseconds since unix epoch as int64 from timestamp message path column
    timestamps_as_ns: "pyarrow.Int64Array"
    if pa.types.is_timestamp(data_type):
        timestamps_as_ns = pc.cast(timestamp_values, pa.timestamp("ns", "UTC"))
        # timestamps are internally stored as 64-bit integers
        # https://arrow.apache.org/docs/python/timestamps.html#arrow-pandas-timestamps
        timestamps_as_ns = typing.cast("pyarrow.Int64Array", timestamps_as_ns.view(pa.int64()))
    elif pa.types.is_floating(data_type):
        multiplier = pa.scalar(timestamp.unit().nano_multiplier())
        # Ensure double precision to avoid overflow
        timestamps_as_doubles = pc.cast(timestamp_values, pa.float64())
        scaled = pc.multiply_checked(timestamps_as_doubles, multiplier)
        timestamps_as_ns = pc.trunc(scaled).cast(pa.int64())
    elif pa.types.is_decimal(data_type):
        # Find narrowest precision decimal that multiplier will fit into.
        # Result of scaling will be a decimal array with precision
        # equal to `sum(timestamp_precision + multipier_precision) + 1`.
        # Without casting multiplier to smallest possible decimal,
        # pyarrow will cast it to the same type as the timestamp values,
        # leading to an error like (e.g.):
        #   > pyarrow.lib.ArrowInvalid: Decimal precision out of range [1, 38]: 39
        nano_multiplier = timestamp.unit().nano_multiplier()
        multiplier_precision = int(math.log10(nano_multiplier)) + 1
        multiplier = pa.scalar(nano_multiplier, pa.decimal64(multiplier_precision, 0))
        scaled = pc.multiply_checked(timestamp_values, multiplier)
        timestamps_as_ns = scaled.cast(pa.int64())
    elif pa.types.is_integer(data_type):
        multiplier = pa.scalar(timestamp.unit().nano_multiplier())
        scaled = pc.multiply_checked(timestamp_values, multiplier)
        timestamps_as_ns = scaled.cast(pa.int64())
    else:
        raise TypeError(
            f"Roboto does not support timestamps formatted as {data_type}. "
            "This is likely an issue with data ingestion. Please contact Roboto support."
        )

    return timestamps_as_ns


def extract_timestamp_field(schema: "pyarrow.Schema", timestamp_message_path: MessagePathRecord) -> Timestamp:
    """Aggregate timestamp info into a helper utility for handling time-based data operations."""
    arrow_field = schema.field(timestamp_message_path.source_path)
    return Timestamp(field=arrow_field, message_path=timestamp_message_path)


def should_read_row_group(
    row_group_metadata: "pyarrow.parquet.RowGroupMetaData",
    timestamp: Timestamp,
    start_time: typing.Optional[int] = None,
    end_time: typing.Optional[int] = None,
) -> bool:
    """
    Determine whether a Parquet row group contains data within the requested time range.
    Used to short-circuit requesting column chunks from the given row group if not relevant.
    """
    for col_idx in range(row_group_metadata.num_columns):
        col_chunk_meta = row_group_metadata.column(col_idx)
        if col_chunk_meta.path_in_schema != timestamp.field.name:
            continue

        stats = col_chunk_meta.statistics
        if stats is None:
            # Without column chunk statistics,
            # we can't know without reading the row group whether it has relevant data
            return True

        max_val = stats.max
        if start_time is not None and max_val is not None and start_time > timestamp.to_epoch_nanoseconds(max_val):
            # Target window of data starts "after" the data contained by this row group
            return False

        min_val = stats.min
        if end_time is not None and min_val is not None and end_time < timestamp.to_epoch_nanoseconds(min_val):
            # Target window of data ends "before" the data contained by this row group
            return False

        return True

    # Couldn't figure out the answer by looking at file metadata, so read the row group
    return True


def _list_ancestor_column(
    schema: "pyarrow.Schema",
    path_in_schema: list[str],
) -> typing.Optional[str]:
    """Return the column path of the nearest list ancestor, or ``None``.

    Walks *path_in_schema* through the Arrow *schema*, checking each intermediate
    field.  If any intermediate field is a list (or large-list) type, returns the
    dot-joined path up to and including that list field — which is the column name
    that should be passed to ``read_row_group(columns=...)`` instead of the full
    leaf path.

    Top-level fields (single-element paths) are never considered nested inside a
    list, so this always returns ``None`` for them.

    Examples::

        # points: list<item: struct<x: int64, y: int64>>
        _list_ancestor_column(schema, ["points", "x"])  # → "points"

        # outer: struct<inner: list<item: struct<x: int64>>>
        _list_ancestor_column(schema, ["outer", "inner", "x"])  # → "outer.inner"

        # position: struct<x: float64, y: float64>
        _list_ancestor_column(schema, ["position", "x"])  # → None
    """
    pa = import_optional_dependency("pyarrow", "analytics")

    if len(path_in_schema) <= 1:
        return None

    current_type = schema.field(path_in_schema[0]).type
    for idx, part in enumerate(path_in_schema[1:], start=1):
        if pa.types.is_list(current_type) or pa.types.is_large_list(current_type):
            return ".".join(path_in_schema[:idx])
        if pa.types.is_struct(current_type):
            current_type = current_type.field(part).type
        else:
            break
    return None


def resolve_columns(
    schema: "pyarrow.Schema",
    message_paths: collections.abc.Iterable[MessagePathRecord],
) -> list[str]:
    """Build a deduplicated list of column names safe for ``read_row_group(columns=...)``.

    Children of list-type columns are replaced by their list ancestor's column name
    because PyArrow's prefix-based nested column selection does not work through list
    wrapper nodes in the physical Parquet schema.  Selecting the parent list column
    already returns its full nested structure.

    This is important because the message-path-to-representation mappings returned by
    the server contain only *leaf* message paths.  For a column like
    ``points: list<struct<x, y>>``, only ``points.x`` and ``points.y`` appear in the
    mapping — the parent ``points`` record is absent.  This function derives the
    correct parent column name from the child's ``path_in_schema``.

    Children of struct-type columns are preserved because PyArrow can resolve them via
    dot-separated prefix matching (e.g. ``"position.x"`` selects the ``x`` child of
    the ``position`` struct).
    """
    columns: list[str] = []
    seen: set[str] = set()
    for mp in message_paths:
        ancestor = _list_ancestor_column(schema, mp.path_in_schema)
        col = ancestor if ancestor is not None else mp.source_path
        if col not in seen:
            columns.append(col)
            seen.add(col)
    return columns



# Copyright (c) 2025 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

import datetime
import typing

import pydantic

from ..time import utcnow
from .object_store import Credentials


class RobotoCredentials(pydantic.BaseModel):
    """
    Credentials returned from the Roboto Platform
    """

    access_key_id: str
    bucket: str
    expiration: datetime.datetime
    secret_access_key: str
    session_token: str
    region: str
    required_prefix: str

    def is_expired(self) -> bool:
        return utcnow() >= self.expiration

    def to_dict(self) -> dict[str, typing.Any]:
        return self.model_dump(mode="json")

    def to_object_store_credentials(self) -> Credentials:
        return {
            "access_key": self.access_key_id,
            "secret_key": self.secret_access_key,
            "token": self.session_token,
            "expiry_time": self.expiration.isoformat(),
            "region": self.region,
        }



# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

from ..command import RobotoCommandSet
from .start import start_command

command_set = RobotoCommandSet(
    name="chat",
    help="Talk with your data, using natural language.",
    commands=[start_command],
)



# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

import typing

import pydantic
from pydantic import ConfigDict

from ...pydantic import (
    remove_non_noneable_init_args,
)
from ...sentinels import NotSet, NotSetType
from ...updates import MetadataChangeset


class CreateDatasetRequest(pydantic.BaseModel):
    """Request payload for creating a new dataset.

    Used to specify the initial properties of a dataset during creation,
    including optional metadata, tags, name, and description.
    """

    description: typing.Optional[str] = pydantic.Field(
        default=None,
        description="An optional human-readable description for this dataset.",
    )
    """Optional human-readable description of the dataset."""

    device_id: typing.Optional[str] = pydantic.Field(
        default=None,
        description="The ID of the device which created this dataset, if applicable.",
    )
    """Optional identifier of the device that generated this data."""

    name: typing.Optional[str] = pydantic.Field(
        default=None,
        description="A short name for this dataset. Must be under 120 characters or less.",
        max_length=120,
    )
    """Optional short name for the dataset (max 120 characters)."""

    metadata: dict[str, typing.Any] = pydantic.Field(
        default_factory=dict,
        description="Initial key-value pairs to associate with this dataset for discovery and search, e.g. "
        + "`{ 'softwareVersion': '3.1.4' }`",
    )
    """Key-value metadata pairs to associate with the dataset for discovery and search."""

    tags: list[str] = pydantic.Field(
        default_factory=list,
        description="Initial tags to associate with this dataset for discovery and search, e.g. "
        + "`['sunny', 'campaign5']`",
    )
    """List of tags for dataset discovery and organization."""

    def __init__(self, **data):
        super().__init__(**remove_non_noneable_init_args(data, self))


class QueryDatasetFilesRequest(pydantic.BaseModel):
    """Request payload for querying files within a dataset.

    Used to retrieve files from a dataset with optional pattern-based filtering
    and pagination support. Supports gitignore-style patterns for flexible
    file selection.
    """

    page_token: typing.Optional[str] = None
    """Token for retrieving the next page of results in paginated queries."""

    include_patterns: typing.Optional[list[str]] = None
    """List of gitignore-style patterns for files to include in results."""

    exclude_patterns: typing.Optional[list[str]] = None
    """List of gitignore-style patterns for files to exclude from results."""

    limit: typing.Optional[int] = None
    """Maximum number of files to return per page."""

    sort_by: typing.Optional[str] = None
    """Field to sort results by. Defaults to 'created'."""

    sort_direction: typing.Optional[str] = None
    """Sort direction ('ASC' or 'DESC'). Defaults to 'DESC'."""


class QueryDatasetsRequest(pydantic.BaseModel):
    """Request payload for querying datasets with filters.

    Used to search for datasets based on various criteria such as metadata,
    tags, and other dataset properties. The filters are applied server-side
    to efficiently return matching datasets.
    """

    filters: dict[str, typing.Any] = pydantic.Field(default_factory=dict)
    """Dictionary of filter criteria to apply when searching for datasets."""

    model_config = ConfigDict(extra="ignore")


class UpdateDatasetRequest(pydantic.BaseModel):
    """Request payload for updating dataset properties.

    Used to modify dataset metadata, description, name, and other properties.
    Only specified fields will be updated; others remain unchanged.
    """

    metadata_changeset: typing.Union[MetadataChangeset, NotSetType] = NotSet
    """Metadata changes to apply (add, update, or remove fields/tags)."""

    description: typing.Optional[typing.Union[str, NotSetType]] = NotSet
    """New description for the dataset. Set to None to clear the description."""

    device_id: typing.Optional[typing.Union[str, NotSetType]] = NotSet
    """New device ID for the dataset. Set to None to clear the device association."""

    name: typing.Optional[
        typing.Union[typing.Annotated[str, pydantic.StringConstraints(max_length=120)], NotSetType]
    ] = NotSet
    """New name for the dataset (max 120 characters). Set to None to clear the name."""

    model_config = pydantic.ConfigDict(extra="ignore", json_schema_extra=NotSetType.openapi_schema_modifier)


class DeleteDirectoriesRequest(pydantic.BaseModel):
    """Request payload for deleting directories within a dataset.

    Used to remove entire directory structures and all contained files
    from a dataset. This is a bulk operation that affects multiple files.
    """

    directory_paths: list[str]
    """List of directory paths to delete from the dataset."""


class RenameDirectoryRequest(pydantic.BaseModel):
    """Request payload for renaming a directory within a dataset.

    Used to change the path of a directory and all its contained files
    within a dataset. This updates the logical organization without
    moving actual file content.
    """

    new_path: str
    """New path for the directory."""

    old_path: str
    """Current path of the directory to rename."""


class CreateDirectoryRequest(pydantic.BaseModel):
    """
    Request payload to create a directory in a dataset
    """

    name: str
    error_if_exists: bool = False
    parent_path: typing.Optional[str] = None
    origination: typing.Optional[str] = None
    create_intermediate_dirs: bool = False
    """If True, creates intermediate directories in the path if they don't exist.
    If False, requires all parent directories to already exist."""


class CreateDatasetIfNotExistsRequest(pydantic.BaseModel):
    """Request payload to create a dataset if no existing dataset matches
    the specified query.

    Searches for existing datasets using the provided RoboQL query. If a matching
    dataset is found, returns that dataset. If no match is found, creates a new
    dataset with the specified properties and returns it.
    """

    match_roboql_query: str
    create_request: CreateDatasetRequest


class BeginSingleFileUploadRequest(pydantic.BaseModel):
    """
    Request payload to begin a single file upload
    """

    origination: typing.Optional[str] = pydantic.Field(
        description="Additional information about what uploaded the file, e.g. `roboto client v1.0.0`.",
        default=None,
    )
    file_path: str = pydantic.Field(
        description=(
            "The destination path to upload the file to in the dataset, e.g. `recording.bag`or `path/to/metadata.json`."
        ),
    )
    file_size: int = pydantic.Field(description="The size of the file in bytes.")


class BeginSingleFileUploadResponse(pydantic.BaseModel):
    """
    Response to a single file upload
    """

    upload_id: str
    upload_url: str



# Copyright (c) 2025 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

from typing import Any, Optional

import pydantic


class RobotoLLMContext(pydantic.BaseModel):
    """Contextual information about what a user is trying to do. May be passed along to Roboto LLM based code paths
    in order to enrich results"""

    dataset_ids: list[str] = pydantic.Field(default_factory=list)
    """IDs of datasets that are relevant to the user's query."""

    file_ids: list[str] = pydantic.Field(default_factory=list)
    """IDs of files that are relevant to the user's query."""

    visualizer_state: Optional[dict[str, Any]] = None
    """State of the visualizer, if a request is being made from the visualizer. This is expected to be a relatively
    opaque JSON blob"""

    misc_context: Optional[dict[str, Any]] = None
    """Miscellaneous context that is relevant to the user's query."""



# Copyright (c) 2025 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

from .api_operations import (
    AbortTransactionsRequest,
    BeginSignedUrlUploadRequest,
    BeginSignedUrlUploadResponse,
    BeginUploadRequest,
    BeginUploadResponse,
    ReportUploadProgressRequest,
)
from .credentials import RobotoCredentials
from .download_session import DownloadableFile
from .file_service import FileService

__all__ = (
    "AbortTransactionsRequest",
    "BeginSignedUrlUploadRequest",
    "BeginSignedUrlUploadResponse",
    "BeginUploadRequest",
    "BeginUploadResponse",
    "DownloadableFile",
    "FileService",
    "ReportUploadProgressRequest",
    "RobotoCredentials",
)



# Copyright (c) 2025 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

"""
Public testing utilities.
"""

from .fake_http_response_factory import FakeHttpResponseFactory
from .stub_roboto_client import StubRobotoClient

__all__ = ["FakeHttpResponseFactory", "StubRobotoClient"]



# Copyright (c) 2025 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

import datetime
import typing
from typing import Any, Optional, Union

import pydantic

from ...compat import StrEnum
from ...time import utcnow
from ..core import RobotoLLMContext


class ChatRole(StrEnum):
    """Enumeration of possible roles in a chat conversation.

    Defines the different participants that can send messages in a chat session.
    """

    USER = "user"
    """Human user sending messages to the AI assistant."""

    ASSISTANT = "assistant"
    """AI assistant responding to user queries and requests."""

    ROBOTO = "roboto"
    """Roboto system providing tool results and system information."""


class ChatMessageStatus(StrEnum):
    """Enumeration of possible message generation states.

    Tracks the lifecycle of message generation from initiation to completion.
    """

    NOT_STARTED = "not_started"
    """Message has been queued but generation has not begun."""

    GENERATING = "generating"
    """Message content is currently being generated."""

    COMPLETED = "completed"
    """Message generation has finished and content is complete."""

    FAILED = "failed"
    """Message generation failed due to an error."""

    CANCELLED = "cancelled"
    """Message generation was cancelled by the user."""

    def is_terminal(self) -> bool:
        """Check if the message generation is in a terminal state.

        Returns:
            True if the message is in a terminal state, False otherwise.
        """
        return self in (ChatMessageStatus.COMPLETED, ChatMessageStatus.FAILED, ChatMessageStatus.CANCELLED)


class ChatStatus(StrEnum):
    """Enumeration of possible chat session states.

    Tracks the overall status of a chat session from creation to termination.
    """

    NOT_STARTED = "not_started"
    """Chat session has been created but no messages have been sent."""

    USER_TURN = "user_turn"
    """User has the turn to send a message."""

    ROBOTO_TURN = "roboto_turn"
    """Roboto is generating a message."""


class ChatContentType(StrEnum):
    """Enumeration of different types of content within chat messages.

    Defines the various content types that can be included in chat messages.
    """

    TEXT = "text"
    """Plain text content from users or AI responses."""

    TOOL_USE = "tool_use"
    """Tool invocation requests from the AI assistant."""

    TOOL_RESULT = "tool_result"
    """Results returned from tool executions."""

    ERROR = "error"
    """Error information when message generation fails."""


class ChatTextContent(pydantic.BaseModel):
    """Text content within a chat message."""

    text: str
    """The actual text content of the message."""

    def __str__(self) -> str:
        return self.text


class ChatToolUseContent(pydantic.BaseModel):
    """Tool usage request content within a chat message."""

    content_type: typing.Literal[ChatContentType.TOOL_USE] = ChatContentType.TOOL_USE
    tool_name: str
    tool_use_id: str
    raw_request: Optional[dict[str, Any]] = None


class ChatToolResultContent(pydantic.BaseModel):
    """Tool execution result content within a chat message."""

    content_type: typing.Literal[ChatContentType.TOOL_RESULT] = ChatContentType.TOOL_RESULT
    tool_name: str
    tool_use_id: str
    runtime_ms: int
    status: str
    raw_response: Optional[dict[str, Any]] = None


class ChatErrorContent(pydantic.BaseModel):
    """Error content within a chat message.

    Used when message generation fails due to an error or is cancelled by the user.
    """

    content_type: typing.Literal[ChatContentType.ERROR] = ChatContentType.ERROR
    error_message: str
    """User-friendly error message describing what went wrong."""

    error_code: Optional[str] = None
    """Optional error code for programmatic handling."""


ChatContent: typing.TypeAlias = Union[ChatTextContent, ChatToolUseContent, ChatToolResultContent, ChatErrorContent]
"""Type alias for all possible content types within chat messages."""


class ChatMessage(pydantic.BaseModel):
    """A single message within a chat conversation.

    Represents one message in the conversation, containing the sender role,
    content blocks, and generation status. Messages can contain multiple
    content blocks of different types (text, tool use, tool results).
    """

    role: ChatRole
    """The role of the message sender (user, assistant, or roboto)."""

    content: list[ChatContent]
    """List of content blocks that make up this message."""

    status: ChatMessageStatus = ChatMessageStatus.NOT_STARTED
    """Current generation status of this message."""

    created: datetime.datetime = pydantic.Field(default_factory=utcnow)
    """Timestamp when this message was created."""

    @classmethod
    def text(cls, text: str, role: ChatRole = ChatRole.USER) -> "ChatMessage":
        """Create a simple text message.

        Convenience method for creating a message containing only text content.

        Args:
            text: The text content for the message.
            role: The role of the message sender. Defaults to USER.

        Returns:
            ChatMessage instance containing the text content.
        """
        return cls(role=role, content=[ChatTextContent(text=text)])

    def __str__(self) -> str:
        combined_content = "".join(str(content) for content in self.content)
        return f"== {self.role.value.upper()} ==\n{combined_content}"

    def is_complete(self) -> bool:
        """Check if message generation is complete.

        Returns:
            True if the message status is COMPLETED, False otherwise.
        """
        return self.status == ChatMessageStatus.COMPLETED

    def is_unsuccessful(self) -> bool:
        """Check if message generation failed or was cancelled.

        Returns:
            True if the message status is FAILED or CANCELLED, False otherwise.
        """
        return self.role != ChatRole.USER and self.status in (ChatMessageStatus.FAILED, ChatMessageStatus.CANCELLED)


class ChatRecord(pydantic.BaseModel):
    """Complete record of a chat session.

    Contains all the persistent data for a chat session including metadata,
    message history, and synchronization state.
    """

    chat_id: str
    """Unique identifier for this chat session."""

    created: datetime.datetime
    """Timestamp when the chat session was created."""

    created_by: str
    """User ID of the person who created this chat session."""

    org_id: str
    """Organization ID that owns this chat session."""

    messages: list[ChatMessage] = pydantic.Field(default_factory=list)
    """Complete list of messages in the conversation."""

    continuation_token: str
    """Token used for incremental updates and synchronization."""

    status: ChatStatus
    """Current status of the chat session."""

    title: Optional[str] = None
    """Title of the chat session."""

    def calculate_new_status(self) -> ChatStatus:
        """Calculate the new status of the chat session based on the latest message."""
        if len(self.messages) == 0:
            return ChatStatus.NOT_STARTED

        latest_message = self.messages[-1]

        # Failed or cancelled messages always return control to the user
        if latest_message.is_unsuccessful():
            return ChatStatus.USER_TURN

        # Normal completion: assistant message with text content
        is_user_turn = (
            latest_message.role == ChatRole.ASSISTANT
            and latest_message.status == ChatMessageStatus.COMPLETED
            and len(latest_message.content) > 0
            and isinstance(latest_message.content[-1], ChatTextContent)
        )

        if is_user_turn:
            return ChatStatus.USER_TURN
        else:
            return ChatStatus.ROBOTO_TURN


class ChatRecordDelta(pydantic.BaseModel):
    """Incremental update to a chat session.

    Contains only the changes since the last synchronization, used for
    efficient real-time updates without transferring the entire chat history.
    """

    messages_by_idx: dict[int, ChatMessage]
    """New or updated messages indexed by their position in the conversation."""

    continuation_token: str
    """Updated token for the next incremental synchronization."""

    status: Optional[ChatStatus] = None
    """Updated status of the chat session."""

    title: Optional[str] = None
    """Updated title of the chat session."""

    def calculate_new_status(self) -> ChatStatus | None:
        """Calculate the new status of the chat session based on the latest message."""
        if len(self.messages_by_idx) == 0:
            return None

        latest_message_idx = max(self.messages_by_idx.keys())
        latest_message = self.messages_by_idx[latest_message_idx]

        # Failed or cancelled messages always return control to the user
        if latest_message.is_unsuccessful():
            return ChatStatus.USER_TURN

        # Normal completion: assistant message with text content
        is_user_turn = (
            latest_message.role == ChatRole.ASSISTANT
            and latest_message.status == ChatMessageStatus.COMPLETED
            and len(latest_message.content) > 0
            and isinstance(latest_message.content[-1], ChatTextContent)
        )

        if is_user_turn:
            return ChatStatus.USER_TURN
        else:
            return ChatStatus.ROBOTO_TURN


class SendMessageRequest(pydantic.BaseModel):
    """Request payload for sending a message to a chat session.

    Contains the message content and optional context for the AI assistant.
    """

    context: Optional[RobotoLLMContext] = None
    """Optional context to include with the message."""

    message: ChatMessage
    """Message content to send."""


class StartChatRequest(pydantic.BaseModel):
    """Request payload for starting a new chat session.

    Contains the initial messages and configuration for creating a new
    chat conversation.
    """

    context: Optional[RobotoLLMContext] = None
    """Optional context to include with the message."""

    messages: list[ChatMessage]
    """Initial messages to start the conversation with."""

    system_prompt: Optional[str] = None
    """Optional system prompt to customize AI assistant behavior."""



# Copyright (c) 2025 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

from __future__ import annotations

import pathlib
import statistics
import typing

from ....compat import import_optional_dependency
from ....exceptions import (
    IngestionException,
    TimestampFieldNotFoundException,
)
from ....logging import default_logger
from ....time import TimeUnit
from .timestamp import (
    TimestampInfo,
    is_timestamp_like,
    is_timezone_aware,
    time_unit_from_timestamp_type,
)

if typing.TYPE_CHECKING:
    import pyarrow  # pants: no-infer-dep
    import pyarrow.parquet  # pants: no-infer-dep


logger = default_logger()


class ParquetParser:
    __file: "pyarrow.parquet.ParquetFile"
    __min_required_row_group_size: int
    __schema: "pyarrow.Schema"
    __small_row_group_count_threshold: int

    @staticmethod
    def is_parquet_file(path: pathlib.Path) -> bool:
        try:
            with path.open("rb") as f:
                magic_bytes = f.read(4)
                return magic_bytes == b"PAR1"
        except (FileNotFoundError, IOError, OSError):
            # If we can't read the file, fall back to extension-based detection
            return path.suffix.lower() == ".parquet"

    def __init__(
        self,
        source: pathlib.Path,
        min_required_row_group_size: int = 100_000,
        small_row_group_count_threshold: int = 32,
    ):
        pq = import_optional_dependency("pyarrow.parquet", "ingestion")

        self.__file = pq.ParquetFile(source)
        self.__min_required_row_group_size = min_required_row_group_size
        self.__schema = self.__file.schema_arrow
        self.__small_row_group_count_threshold = small_row_group_count_threshold

    @property
    def column_count(self) -> int:
        return self.__file.metadata.num_columns

    @property
    def fields(self) -> typing.Generator["pyarrow.Field", None, None]:
        yield from self.__schema

    @property
    def row_count(self) -> int:
        return self.__file.metadata.num_rows

    @property
    def row_group_count(self) -> int:
        return self.__file.metadata.num_row_groups

    @property
    def row_group_size(self) -> int:
        if self.__file.metadata.num_row_groups == 0:
            return 0

        if self.__file.metadata.num_row_groups == 1:
            return self.__file.metadata.row_group(0).num_rows

        # avg row group sizes, omitting the final row group from the avg (not expected to be full)
        row_group_counts = [
            self.__file.metadata.row_group(i).num_rows for i in range(self.__file.metadata.num_row_groups - 1)
        ]
        return round(statistics.mean(row_group_counts))

    def find_timestamp_field_by_type(self) -> "pyarrow.Field":
        pa = import_optional_dependency("pyarrow", "ingestion")

        for field in self.__schema:
            if pa.types.is_timestamp(field.type):
                if is_timezone_aware(field.type):
                    return field
                logger.warning(
                    "'%s' is timestamp-like but is not timezone aware "
                    "and therefore cannot be treated as a timestamp by Roboto",
                    field.name,
                )

        raise TimestampFieldNotFoundException(
            "Unable to determine column that should be treated as the timestamp. "
            "Try providing the timestamp column explicitly, ensuring that column exists in the data, "
            "and is timezone-aware if timestamp-like (e.g., a datetime instance)."
        )

    def extract_timestamp_info(
        self,
        timestamp_column_name: typing.Optional[str] = None,
        timestamp_unit: typing.Optional[typing.Union[str, TimeUnit]] = None,
    ) -> TimestampInfo:
        pa = import_optional_dependency("pyarrow", "ingestion")
        pc = import_optional_dependency("pyarrow.compute", "ingestion")

        field = (
            self.get_timestamp_field_by_name(timestamp_column_name)
            if timestamp_column_name is not None
            else self.find_timestamp_field_by_type()
        )

        unit = None
        if timestamp_unit is not None:
            unit = TimeUnit(timestamp_unit)

        inferred_unit = time_unit_from_timestamp_type(field.type) if pa.types.is_timestamp(field.type) else None

        if unit is None:
            unit = inferred_unit

        if unit is not None and inferred_unit is not None and unit != inferred_unit:
            logger.warning(
                "Timestamp unit provided explicitly but data type suggests '%s'. "
                "Using the explicitly provided unit '%s' instead.",
                inferred_unit.value,
                unit.value,
            )

        if unit is None:
            raise IngestionException(
                f"The timestamp unit cannot be determined for field '{field.name}.' "
                "Explicitly set the timestamp unit or ensure the field is datetime-like."
            )

        table = self.get_data_for_column(field.name)
        data = table.column(0)
        min_max = pc.min_max(data)
        # If the value is in nanoseconds, we'll get an error like
        #
        # ValueError: Nanosecond resolution temporal type 999 is not safely convertible to microseconds to convert
        # to datetime.datetime. Install pandas to return as Timestamp with nanosecond support or
        # access the .value attribute.
        #
        # ...which we can handle by just grabbing the int value as "value". We'll normalize it later anyway.
        try:
            start_time = min_max["min"].as_py()
            end_time = min_max["max"].as_py()
        except ValueError:
            start_time = min_max["min"].value  # type: ignore
            end_time = min_max["max"].value  # type: ignore

        return TimestampInfo(
            field=field,
            unit=unit,
            start_time=start_time,
            end_time=end_time,
        )

    def get_data_for_column(self, column_name: str) -> "pyarrow.Table":
        return self.__file.read(columns=[column_name])

    def get_timestamp_field_by_name(self, column_name: str) -> "pyarrow.Field":
        pa = import_optional_dependency("pyarrow", "ingestion")

        try:
            field = self.__schema.field(column_name)
        except KeyError:
            raise TimestampFieldNotFoundException(
                f"'{column_name}' provided as timestamp, but that field is not present in the provided data.",
            ) from None

        # Does it look like a timestamp?
        if not is_timestamp_like(field.type):
            raise IngestionException(
                f"'{column_name}' provided as timestamp, "
                f"but it is of type '{field.type}'. Expected a datetime-like or numeric field.",
            )

        if pa.types.is_timestamp(field.type) and not is_timezone_aware(field.type):
            raise IngestionException(
                f"'{column_name}' provided as timestamp and is of the proper type, "
                "but is not timezone aware and therefore cannot be treated as a timestamp by Roboto."
            )

        return field

    def requires_rewrite(self, timestamp: TimestampInfo) -> bool:
        if self.row_count == 0:
            # Edge case: the file has no data, no need to rewrite
            return False

        # Parquet files with many small row groups will perform poorly in Roboto.
        many_row_groups = self.row_group_count > self.__small_row_group_count_threshold
        row_groups_are_small = self.row_group_size < self.__min_required_row_group_size
        many_small_row_groups = many_row_groups and row_groups_are_small
        if many_small_row_groups:
            logger.warning(
                "This file must be rewritten because it has many small row groups, "
                "which will cause it to integrate poorly with the Roboto Platform."
            )
            return True

        # Selecting slices of files depends on row group filtering based on timestamp statistics.
        # Rewrite the file if the timestamp column is missing statistics.
        missing_row_group_stats = self.__is_timestamp_column_missing_stats(timestamp.field)
        if missing_row_group_stats:
            logger.warning(
                "This file must be rewritten because it has missing row group statistics for the timestamp column, "
                "which will cause it to integrate poorly with the Roboto Platform."
            )
            return True

        return False

    def rewrite(
        self,
        outfile: pathlib.Path,
        timestamp: TimestampInfo,
        # The bigger the better, but there is a direct correlation between this and system memory utilization
        # This can (and maybe should) be dialed up to 500MB or 1GB, but that would require a beefier VM.
        target_row_group_size_bytes: int = 100 * 1000 * 1000,  # 100MB
    ) -> None:
        pq = import_optional_dependency("pyarrow.parquet", "ingestion")

        metadata = self.__file.metadata

        # determine optimal row group size to use while reading the source file to:
        #   1. make sure this action doesn't run out of memory
        #   2. ensure read efficiency of the target file
        #       (as currently implemented, batch size at read time is the row group size at write time)
        max_bytes_per_row = (
            max(
                (metadata.row_group(i).total_byte_size / metadata.row_group(i).num_rows)
                for i in range(metadata.num_row_groups)
                if metadata.row_group(i).num_rows > 0
            )
            if metadata.num_row_groups > 0
            else 1000
        )

        row_group_size = max(int(target_row_group_size_bytes / max_bytes_per_row), 100_000)

        logger.info(
            "Using row group size of %d rows based on an estimated %d bytes per row",
            row_group_size,
            max_bytes_per_row,
        )
        with pq.ParquetWriter(
            where=outfile,
            schema=self.__schema,
            # https://github.com/apache/parquet-format/blob/eb4b31c1d64a01088d02a2f9aefc6c17c54cc6fc/src/main/thrift/parquet.thrift#L619
            # vs
            # https://github.com/apache/parquet-format/blob/eb4b31c1d64a01088d02a2f9aefc6c17c54cc6fc/src/main/thrift/parquet.thrift#L577
            # Supported by our in-browser reader:
            # https://github.com/hyparam/hyparquet/blob/master/src/datapage.js
            data_page_version="2.0",
            # Statistics are used for row group filtering
            write_statistics=[timestamp.field.name],
            # https://arrow.apache.org/docs/python/parquet.html#storing-timestamps
            version="2.6",
        ) as writer:
            for batch in self.__file.iter_batches(batch_size=row_group_size):
                writer.write_batch(batch)

    def __is_timestamp_column_missing_stats(self, timestamp_field: "pyarrow.Field") -> bool:
        for row_group_idx in range(self.__file.metadata.num_row_groups):
            row_group = self.__file.metadata.row_group(row_group_idx)
            for col_idx in range(self.__file.metadata.num_columns):
                column = row_group.column(col_idx)
                if column.path_in_schema != timestamp_field.name:
                    continue
                if not column.is_stats_set:
                    return True
        return False



# Copyright (c) 2025 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

from __future__ import annotations

import collections.abc
import dataclasses
import email.message
import io
import json
import typing
import urllib.response

import pydantic

from roboto.http import RobotoClient
from roboto.http.response import HttpResponse

# Type alias for the response
StubResponse = typing.Union[
    dict[str, typing.Any],
    list[typing.Any],
    str,
    int,
    float,
    bool,
    None,
    pydantic.BaseModel,
    Exception,
]

RequestBodyMatcher = typing.Callable[[typing.Any], bool]


@dataclasses.dataclass
class _Expectation:
    """Internal representation of an expected request/response pair."""

    method: str
    path: str
    response: StubResponse
    query: typing.Optional[dict[str, typing.Any]] = None
    request_body_matcher: typing.Optional[RequestBodyMatcher] = None


class StubRobotoClient(RobotoClient):
    """
    A stub RobotoClient for unit testing that returns pre-configured responses for expected requests.

    Example:
        >>> from roboto_test_utils import StubRobotoClient
        >>> from roboto.domain.datasets import Dataset, DatasetRecord
        >>> from roboto.time import utcnow
        >>>
        >>> client = StubRobotoClient()
        >>> client.expect_get(
        ...     "/v1/datasets/ds_123",
        ...     response=DatasetRecord(
        ...         dataset_id="ds_123",
        ...         created=utcnow(),
        ...         created_by="test@example.com",
        ...         modified=utcnow(),
        ...         modified_by="test@example.com",
        ...         org_id="og_test",
        ...     ),
        ... )
        >>>
        >>> dataset = Dataset.from_id("ds_123", roboto_client=client)
        >>> assert dataset.dataset_id == "ds_123"
    """

    __expectations: list[_Expectation]

    def __init__(self) -> None:
        super().__init__(endpoint="https://stub.roboto.test", auth_decorator=None)
        self.__expectations = []

    def __enter__(self) -> StubRobotoClient:
        return self

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        self.verify_no_pending_expectations()

    def expect_get(
        self,
        path: str,
        response: StubResponse,
        query: typing.Optional[dict[str, typing.Any]] = None,
    ) -> StubRobotoClient:
        self.__expectations.append(_Expectation("GET", path, response, query))
        return self

    def expect_post(
        self,
        path: str,
        response: StubResponse,
        query: typing.Optional[dict[str, typing.Any]] = None,
        request_body_matcher: typing.Optional[RequestBodyMatcher] = None,
    ) -> StubRobotoClient:
        self.__expectations.append(_Expectation("POST", path, response, query, request_body_matcher))
        return self

    def expect_put(
        self,
        path: str,
        response: StubResponse,
        query: typing.Optional[dict[str, typing.Any]] = None,
        request_body_matcher: typing.Optional[RequestBodyMatcher] = None,
    ) -> StubRobotoClient:
        self.__expectations.append(_Expectation("PUT", path, response, query, request_body_matcher))
        return self

    def expect_patch(
        self,
        path: str,
        response: StubResponse,
        query: typing.Optional[dict[str, typing.Any]] = None,
        request_body_matcher: typing.Optional[RequestBodyMatcher] = None,
    ) -> StubRobotoClient:
        self.__expectations.append(_Expectation("PATCH", path, response, query, request_body_matcher))
        return self

    def expect_delete(
        self,
        path: str,
        response: StubResponse,
        query: typing.Optional[dict[str, typing.Any]] = None,
        request_body_matcher: typing.Optional[RequestBodyMatcher] = None,
    ) -> StubRobotoClient:
        self.__expectations.append(_Expectation("DELETE", path, response, query, request_body_matcher))
        return self

    # Override HTTP methods to use expectations
    def get(
        self,
        path: collections.abc.Sequence[str] | str,
        caller_org_id: typing.Optional[str] = None,
        headers: typing.Optional[dict[str, str]] = None,
        idempotent: bool = True,
        owner_org_id: typing.Optional[str] = None,
        query: typing.Optional[dict[str, typing.Any]] = None,
        retry_wait_fn: typing.Optional[typing.Any] = None,
        timeout: typing.Any = None,
    ) -> HttpResponse:
        return self.__consume_expectation("GET", path, query, None)

    def post(
        self,
        path: collections.abc.Sequence[str] | str,
        caller_org_id: typing.Optional[str] = None,
        data: typing.Any = None,
        headers: typing.Optional[dict[str, str]] = None,
        idempotent: bool = True,
        owner_org_id: typing.Optional[str] = None,
        query: typing.Optional[dict[str, typing.Any]] = None,
        retry_wait_fn: typing.Optional[typing.Any] = None,
        timeout: typing.Any = None,
    ) -> HttpResponse:
        return self.__consume_expectation("POST", path, query, data)

    def put(
        self,
        path: collections.abc.Sequence[str] | str,
        caller_org_id: typing.Optional[str] = None,
        data: typing.Any = None,
        headers: typing.Optional[dict[str, str]] = None,
        idempotent: bool = True,
        owner_org_id: typing.Optional[str] = None,
        query: typing.Optional[dict[str, typing.Any]] = None,
        retry_wait_fn: typing.Optional[typing.Any] = None,
        timeout: typing.Any = None,
    ) -> HttpResponse:
        return self.__consume_expectation("PUT", path, query, data)

    def patch(
        self,
        path: collections.abc.Sequence[str] | str,
        caller_org_id: typing.Optional[str] = None,
        data: typing.Any = None,
        headers: typing.Optional[dict[str, str]] = None,
        idempotent: bool = True,
        owner_org_id: typing.Optional[str] = None,
        query: typing.Optional[dict[str, typing.Any]] = None,
        retry_wait_fn: typing.Optional[typing.Any] = None,
        timeout: typing.Any = None,
    ) -> HttpResponse:
        return self.__consume_expectation("PATCH", path, query, data)

    def delete(
        self,
        path: collections.abc.Sequence[str] | str,
        caller_org_id: typing.Optional[str] = None,
        data: typing.Any = None,
        headers: typing.Optional[dict[str, str]] = None,
        idempotent: bool = True,
        owner_org_id: typing.Optional[str] = None,
        query: typing.Optional[dict[str, typing.Any]] = None,
        retry_wait_fn: typing.Optional[typing.Any] = None,
        timeout: typing.Any = None,
    ) -> HttpResponse:
        return self.__consume_expectation("DELETE", path, query, data)

    def verify_no_pending_expectations(self) -> None:
        """
        Verify that all registered expectations have been consumed.

        Raises:
            AssertionError: If there are any pending expectations that were not consumed.
        """
        if self.__expectations:
            pending = [(e.method, e.path) for e in self.__expectations]
            raise AssertionError(f"Pending expectations were not consumed: {pending}")

    def __consume_expectation(
        self,
        method: str,
        path: collections.abc.Sequence[str] | str,
        query: typing.Optional[dict[str, typing.Any]],
        body: typing.Any,
    ) -> HttpResponse:
        """Find and consume a matching expectation, returning its response."""
        normalized_path = self.__normalize_path(path)

        for i, expectation in enumerate(self.__expectations):
            if expectation.method != method:
                continue
            if self.__normalize_path(expectation.path) != normalized_path:
                continue
            if expectation.query is not None and expectation.query != query:
                continue
            if expectation.request_body_matcher and not expectation.request_body_matcher(body):
                continue

            # Found a match - consume it
            self.__expectations.pop(i)

            # If response is an exception, raise it
            if isinstance(expectation.response, Exception):
                raise expectation.response

            return self.__build_response(expectation.response)

        raise AssertionError(
            f"Unexpected request: {method} {normalized_path}\n"
            f"Query: {query}\n"
            f"Body: {body}\n"
            f"Pending expectations: {[(e.method, e.path) for e in self.__expectations]}"
        )

    def __normalize_path(self, path: typing.Union[str, collections.abc.Iterable[str]]) -> str:
        """Normalize path to a string with leading slash."""
        if isinstance(path, str):
            normalized = path
        else:
            normalized = "/".join(path)
        return f"/{normalized.lstrip('/')}"

    def __build_response(self, response: StubResponse) -> HttpResponse:
        """Build an HttpResponse from the stub response data."""
        # Serialize pydantic models
        response_data: typing.Any
        if isinstance(response, pydantic.BaseModel):
            response_data = response.model_dump(mode="json")
        else:
            response_data = response

        # Wrap in {"data": ...}
        wrapped: dict[str, typing.Any] = {"data": response_data}

        # Build the HttpResponse
        headers = email.message.Message()
        headers.add_header("Content-Type", "application/json")
        data = io.BytesIO(json.dumps(wrapped).encode())
        urllib_response = urllib.response.addinfourl(data, headers, "https://stub.roboto.test", 200)
        return HttpResponse(urllib_response)



# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

import logging

LOGGER_NAME = "roboto"


def default_logger():
    return logging.getLogger(LOGGER_NAME)


def maybe_pluralize(word: str, count: int) -> str:
    return word if count == 1 else f"{word}s"



# Copyright (c) 2025 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

from __future__ import annotations

import collections
import collections.abc
import pathlib
import tempfile
import typing

from ...association import Association
from ...domain.actions import InvocationInput
from ...domain.files import File
from ...domain.topics import Topic
from ...fs import DownloadableFile, FileService
from ...http import RobotoClient
from ...logging import default_logger, maybe_pluralize
from ...progress import NoopProgressMonitor, TqdmProgressMonitor
from ...roboto_search import RobotoSearch
from .action_input import ActionInputRecord
from .file_resolver import InputFileResolver
from .topic_resolver import InputTopicResolver

log = default_logger()

OptionallyDownloadedFile: typing.TypeAlias = tuple[File, pathlib.Path | None]


class ActionInputResolver:
    """
    Resolves the action invocation input spec to concrete Roboto entities.

    The entities are packaged together in an `ActionInput` instance,
    which is available to action code via `ActionRuntime`.
    """

    @classmethod
    def from_env(
        cls,
        roboto_client: typing.Optional[RobotoClient] = None,
        roboto_search: typing.Optional[RobotoSearch] = None,
    ) -> ActionInputResolver:
        roboto_client = RobotoClient.defaulted(roboto_client)
        roboto_search = roboto_search if roboto_search is not None else RobotoSearch.for_roboto_client(roboto_client)

        return cls(
            file_resolver=InputFileResolver(roboto_client, roboto_search),
            topic_resolver=InputTopicResolver(roboto_client, roboto_search),
            file_service=FileService(roboto_client=roboto_client),
        )

    def __init__(
        self,
        file_resolver: InputFileResolver,
        topic_resolver: InputTopicResolver,
        file_service: FileService,
    ):
        self.__file_resolver = file_resolver
        self.__topic_resolver = topic_resolver
        self.__file_service = file_service

    def resolve_input_spec(
        self,
        input_spec: InvocationInput,
        download: bool = False,
        download_path: typing.Optional[pathlib.Path] = None,
    ) -> ActionInputRecord:
        """This method takes an InvocationInput containing data selectors (e.g., for files and topics)
        and resolves them to concrete entities.

        Optionally downloads files to a local path.

        Args:
            input_spec: Input specification containing data selectors.
                See :py:class:`~roboto.domain.actions.InvocationInput` for more detail.
            download: If True, download all resolved files to local disk. Defaults to False.
            download_path: Directory path where files should be downloaded.
                If not provided and download=True, a temporary directory will be created.
                Ignored if download=False.

        Returns:
            ActionInputRecord containing:
            - files: List of (FileRecord, Optional[Path]) tuples. Path is None if download=False,
              otherwise contains the local path where the file was downloaded.
            - topics: List of TopicRecord instances.

        Examples:
            Resolve files using a RoboQL query without downloading:

            >>> input_spec = InvocationInput.file_query('dataset_id = "ds_abc123" AND path LIKE "%.mcap"')
            >>> result = resolver.resolve_input_spec(input_spec)
            >>> # result.files contains (FileRecord, None) tuples
            >>> # result.topics is empty

            Resolve and download files to a specific directory:

            >>> input_spec = InvocationInput.file_query('dataset_id = "ds_abc123" AND path LIKE "%.mcap"')
            >>> result = resolver.resolve_input_spec(input_spec, download=True, download_path=Path("/tmp/data"))
            >>> # result.files contains (FileRecord, Path) tuples with local paths

            Resolve both files and topics:

            >>> input_spec = InvocationInput(
            ...     files=FileSelector(query='dataset_id = "ds_abc123" AND path LIKE "%.mcap"'),
            ...     topics=DataSelector(names=["battery_status", "gps"]),
            ... )
            >>> result = resolver.resolve_input_spec(input_spec)
            >>> # result.files contains file records
            >>> # result.topics contains topic records
        """
        files: list[File] = []
        topics: list[Topic] = []

        if input_spec.files:
            files = self.__file_resolver.resolve_all(input_spec.safe_files)
            if not files:
                log.warning("No files matched the provided input specification.")

        if input_spec.topics:
            topics = self.__topic_resolver.resolve_all(input_spec.safe_topics)
            if not topics:
                log.warning("No topics matched the provided input specification.")

        resolved_files: collections.abc.Sequence[OptionallyDownloadedFile] = [(file, None) for file in files]
        if download:
            if download_path is None:
                download_path = pathlib.Path(tempfile.mkdtemp())
            log.info("Downloading {%d} file(s) ...", len(files))
            resolved_files = self.__download_files(download_path, files)

            if topics:
                log.warning(
                    "Topic data cannot be downloaded during action invocation setup. "
                    "Use topic.get_data() or topic.get_data_as_df() in your action code to access topic data. "
                    "See: https://docs.roboto.ai/reference/python-sdk/roboto/domain/topics/topic/index.html"
                )

        return ActionInputRecord(
            files=[(file.record, path) for file, path in resolved_files],
            topics=[topic.record for topic in topics],
        )

    def __download_files(
        self,
        out_path: pathlib.Path,
        files: collections.abc.Iterable[File],
    ) -> list[tuple[File, pathlib.Path]]:
        """
        Downloads the specified files to the provided local directory.

        Files already present locally with matching size are skipped.

        Args:
            out_path: Destination directory for the downloaded files.
            files: Files to download.

        Returns:
            A list of (File, Path) tuples, relating each provided file to its download path.
        """

        class DownloadCandidate(typing.NamedTuple):
            file: File
            path: pathlib.Path
            requires_download: bool

        download_candidates_by_association: dict[str, list[DownloadCandidate]] = collections.defaultdict(list)

        for file in files:
            local_path = out_path / file.file_id / str(file.version) / pathlib.Path(file.relative_path).name

            # Skip download if file already exists locally with matching size
            requires_download = True
            if local_path.exists():
                local_file_size_matches_remote = local_path.stat().st_size == file.record.size
                if local_file_size_matches_remote:
                    log.debug(
                        "%s (%s@v%d) already downloaded.",
                        file.relative_path,
                        file.file_id,
                        file.version,
                    )
                    requires_download = False

            download_candidates_by_association[file.dataset_id].append(
                DownloadCandidate(file=file, path=local_path, requires_download=requires_download)
            )

        # Calculate total size of files that need downloading
        total_size = sum(
            candidate.file.record.size
            for candidates in download_candidates_by_association.values()
            for candidate in candidates
            if candidate.requires_download
        )
        file_count = sum(
            1
            for candidates in download_candidates_by_association.values()
            for candidate in candidates
            if candidate.requires_download
        )

        progress_monitor = (
            TqdmProgressMonitor(
                total=total_size,
                desc=f"Downloading {file_count} {maybe_pluralize('file', file_count)}",
            )
            if file_count > 0
            else NoopProgressMonitor()
        )

        with progress_monitor:
            for association_id, candidates in download_candidates_by_association.items():
                downloadable: list[DownloadableFile] = [
                    {
                        "bucket_name": candidate.file.record.bucket,
                        "source_uri": candidate.file.record.uri,
                        "destination_path": candidate.path,
                    }
                    for candidate in candidates
                    if candidate.requires_download
                ]

                if downloadable:
                    self.__file_service.download(
                        files=downloadable,
                        association=Association.dataset(association_id),
                        on_progress=progress_monitor.update,
                    )

        return [
            (candidate.file, candidate.path)
            for candidates in download_candidates_by_association.values()
            for candidate in candidates
        ]
